// https://w3.org/TR/WGSL/#grammar
export const WGSL_KEYWORDS = [
  // 5.2.6 Vector Types
  'vec2',
  'vec2i',
  'vec2u',
  'vec2f',
  'vec2h',
  'vec3',
  'vec3i',
  'vec3u',
  'vec3f',
  'vec3h',
  'vec4',
  'vec4i',
  'vec4u',
  'vec4f',
  'vec4h',

  // 5.2.7 Matrix Types
  'mat2x2',
  'mat2x2f',
  'mat2x2h',
  'mat2x3',
  'mat2x3f',
  'mat2x3h',
  'mat2x4',
  'mat2x4f',
  'mat2x4h',
  'mat3x2',
  'mat3x2f',
  'mat3x2h',
  'mat3x3',
  'mat3x3f',
  'mat3x3h',
  'mat3x4',
  'mat3x4f',
  'mat3x4h',
  'mat4x2',
  'mat4x2f',
  'mat4x2h',
  'mat4x3',
  'mat4x3f',
  'mat4x3h',
  'mat4x4',
  'mat4x4f',
  'mat4x4h',

  // 15.1.1 Type-defining Keywords
  'array',
  'atomic',
  'bool',
  'f32',
  'f16',
  'i32',
  // 'mat2x2',
  // 'mat2x3',
  // 'mat2x4',
  // 'mat3x2',
  // 'mat3x3',
  // 'mat3x4',
  // 'mat4x2',
  // 'mat4x3',
  // 'mat4x4',
  'ptr',
  'sampler',
  'sampler_comparison',
  'texture_1d',
  'texture_2d',
  'texture_2d_array',
  'texture_3d',
  'texture_cube',
  'texture_cube_array',
  'texture_multisampled_2d',
  'texture_storage_1d',
  'texture_storage_2d',
  'texture_storage_2d_array',
  'texture_storage_3d',
  'texture_depth_2d',
  'texture_depth_2d_array',
  'texture_depth_cube',
  'texture_depth_cube_array',
  'texture_depth_multisampled_2d',
  'u32',
  // 'vec2',
  // 'vec3',
  // 'vec4',

  // 15.1.2 Other Keywords
  'alias',
  'bitcast',
  'break',
  'case',
  'const',
  'const_assert',
  'continue',
  'continuing',
  'default',
  'discard',
  'else',
  'enable',
  'false',
  'fn',
  'for',
  'if',
  'let',
  'loop',
  'override',
  'return',
  'struct',
  'switch',
  'true',
  'var',
  'while',

  // 15.2 Reserved Words
  'Hullshader',
  'NULL',
  'Self',
  'abstract',
  'active',
  'alignas',
  'alignof',
  'as',
  'asm',
  'asm_fragment',
  'async',
  'attribute',
  'auto',
  'await',
  'become',
  'bf16',
  'binding_array',
  'cast',
  'catch',
  'class',
  'co_await',
  'co_return',
  'co_yield',
  'coherent',
  'column_major',
  'common',
  'compile',
  'compile_fragment',
  'concept',
  'const_cast',
  'consteval',
  'constexpr',
  'constinit',
  'crate',
  'debugger',
  'decltype',
  'delete',
  'demote',
  'demote_to_helper',
  'do',
  'dynamic_cast',
  'enum',
  'explicit',
  'export',
  'extends',
  'extern',
  'external',
  'f64',
  'fallthrough',
  'filter',
  'final',
  'finally',
  'friend',
  'from',
  'fxgroup',
  'get',
  'goto',
  'groupshared',
  'handle',
  'highp',
  'i16',
  'i64',
  'i8',
  'impl',
  'implements',
  'import',
  'inline',
  'instanceof',
  'interface',
  'layout',
  'lowp',
  'macro',
  'macro_rules',
  'match',
  'mediump',
  'meta',
  'mod',
  'module',
  'move',
  'mut',
  'mutable',
  'namespace',
  'new',
  'nil',
  'noexcept',
  'noinline',
  'nointerpolation',
  'noperspective',
  'null',
  'nullptr',
  'of',
  'operator',
  'package',
  'packoffset',
  'partition',
  'pass',
  'patch',
  'pixelfragment',
  'precise',
  'precision',
  'premerge',
  'priv',
  'protected',
  'pub',
  'public',
  'quat',
  'readonly',
  'ref',
  'regardless',
  'register',
  'reinterpret_cast',
  'requires',
  'resource',
  'restrict',
  'self',
  'set',
  'shared',
  'sizeof',
  'smooth',
  'snorm',
  'static',
  'static_assert',
  'static_cast',
  'std',
  'subroutine',
  'super',
  'target',
  'template',
  'this',
  'thread_local',
  'throw',
  'trait',
  'try',
  'type',
  'typedef',
  'typeid',
  'typename',
  'typeof',
  'u16',
  'u64',
  'u8',
  'union',
  'unless',
  'unorm',
  'unsafe',
  'unsized',
  'use',
  'using',
  'varying',
  'virtual',
  'volatile',
  'wgsl',
  'where',
  'with',
  'writeonly',
  'yield',

  // 15.4 Context-Dependent Name Tokens - Attributes
  '@align',
  '@binding',
  '@builtin',
  '@compute',
  '@const',
  '@fragment',
  '@group',
  '@id',
  '@interpolate',
  '@invariant',
  '@location',
  '@size',
  '@vertex',
  '@workgroup_size',

  // 15.4 Context-Dependent Name Tokens - Interpolation Types
  'perspective',
  'linear',
  'flat',

  // 15.4 Context-Dependent Name Tokens - Interpolation Sampling
  'center',
  'centroid',
  'sample',

  // 15.4 Context-Dependent Name Tokens - Built-in Values
  'vertex_index',
  'instance_index',
  'position',
  'front_facing',
  'frag_depth',
  'local_invocation_id',
  'local_invocation_index',
  'global_invocation_id',
  'workgroup_id',
  'num_workgroups',
  'sample_index',
  'sample_mask',

  // 15.4 Context-Dependent Name Tokens - Access Modes
  'read',
  'write',
  'read_write',

  // 15.4 Context-Dependent Name Tokens - Address Spaces
  'function',
  'private',
  'workgroup',
  'uniform',
  'storage',

  // 15.4 Context-Dependent Name Tokens - Texel Formats
  'rgba8unorm',
  'rgba8snorm',
  'rgba8uint',
  'rgba8sint',
  'rgba16uint',
  'rgba16sint',
  'rgba16float',
  'r32uint',
  'r32sint',
  'r32float',
  'rg32uint',
  'rg32sint',
  'rg32float',
  'rgba32uint',
  'rgba32sint',
  'rgba32float',
  'bgra8unorm',

  // 15.4 Context-Dependent Name Tokens - Extensions
  'f16', // v1 optional

  // 17.1 Logical Built-in Functions
  'all',
  'any',
  'select',

  // 17.2 Array Built-in Functions
  'arrayLength',

  // 17.3 Numeric Built-in Functions
  'abs',
  'acos',
  'acosh',
  'asin',
  'asinh',
  'atan',
  'atanh',
  'atan2',
  'ceil',
  'clamp',
  'cos',
  'cosh',
  'countLeadingZeros',
  'countOneBits',
  'countTrailingZeros',
  'cross',
  'degrees',
  'determinant',
  'distance',
  'dot',
  'exp',
  'exp2',
  'extractBits',
  'faceForward',
  'firstLeadingBit',
  'firstTrailingBit',
  'floor',
  'fma',
  'fract',
  'frexp',
  'insertBits',
  'inverseSqrt',
  'ldexp',
  'length',
  'log',
  'log2',
  'max',
  'min',
  'mix',
  'modf',
  'normalize',
  'pow',
  'quantizeToF16',
  'radians',
  'reflect',
  'refract',
  'reverseBits',
  'round',
  'saturate',
  'sign',
  'sin',
  'sinh',
  'smoothstep',
  'sqrt',
  'step',
  'tan',
  'tanh',
  'transpose',
  'trunc',

  // 17.4 Derivative Built-in Functions
  'dpdx',
  'dpdxCoarse',
  'dpdxFine',
  'dpdy',
  'dpdyCoarse',
  'dpdyFine',
  'fwidth',
  'fwidthCoarse',
  'fwidthFine',

  // 17.5 Texture Built-in Functions
  'textureDimensions',
  'textureGather',
  'textureGatherCompare',
  'textureLoad',
  'textureNumLayers',
  'textureNumLevels',
  'textureNumSamples',
  'textureSample',
  'textureSampleBias',
  'textureSampleCompare',
  'textureSampleCompareLevel',
  'textureSampleGrad',
  'textureSampleLevel',
  'textureSampleBaseClampToEdge',

  // 17.6 Atomic Built-in Functions
  'atomicLoad',
  'atomicStore',
  'atomicAdd',
  'atomicSub',
  'atomicMax',
  'atomicMin',
  'atomicAnd',
  'atomicOr',
  'atomicXor',
  'atomicExchange',
  'atomicCompareExchangeWeak',

  // 17.7 Data Packing Built-in Functions
  'pack4x8snorm',
  'pack4x8unorm',
  'pack2x16snorm',
  'pack2x16unorm',
  'pack2x16float',

  // 17.8 Data Unpacking Built-in Functions
  'unpack4x8snorm',
  'unpack4x8unorm',
  'unpack2x16snorm',
  'unpack2x16unorm',
  'unpack2x16float',

  // 17.9 Synchronization Built-in Functions
  'storageBarrier',
  'workgroupBarrier',
  'workgroupUniformLoad',
]

// https://registry.khronos.org/OpenGL/specs/es/3.0/GLSL_ES_Specification_3.00.pdf
// NOTE: restrictions from 5.25-5.26 apply https://registry.khronos.org/webgl/specs/latest/2.0
export const GLSL_KEYWORDS = [
  // 3.8 Keywords
  'const',
  'uniform',
  'layout',
  'centroid',
  'flat',
  'smooth',
  'break',
  'continue',
  'do',
  'for',
  'while',
  'switch',
  'case',
  'default',
  'if',
  'else',
  'in',
  'out',
  'inout',
  'float',
  'int',
  'void',
  'bool',
  'true',
  'false',
  'invariant',
  'discard',
  'return',
  'mat2',
  'mat3',
  'mat4',
  'mat2x2',
  'mat2x3',
  'mat2x4',
  'mat3x2',
  'mat3x3',
  'mat3x4',
  'mat4x2',
  'mat4x3',
  'mat4x4',
  'vec2',
  'vec3',
  'vec4',
  'ivec2',
  'ivec3',
  'ivec4',
  'bvec2',
  'bvec3',
  'bvec4',
  'uint',
  'uvec2',
  'uvec3',
  'uvec4',
  'lowp',
  'mediump',
  'highp',
  'precision',
  'sampler2D',
  'sampler3D',
  'samplerCube',
  'sampler2DShadow',
  'samplerCubeShadow',
  'sampler2DArray',
  'sampler2DArrayShadow',
  'isampler2D',
  'isampler3D',
  'isamplerCube',
  'isampler2DArray',
  'usampler2D',
  'usampler3D',
  'usamplerCube',
  'usampler2DArray',
  'struct',

  // 3.8 Keywords - Reserved for future use
  'attribute',
  'varying',
  'coherent',
  'volatile',
  'restrict',
  'readonly',
  'writeonly',
  'resource',
  'atomic_uint',
  'noperspective',
  'patch',
  'sample',
  'subroutine',
  'common',
  'partition',
  'active',
  'asm',
  'class',
  'union',
  'enum',
  'typedef',
  'template',
  'this',
  'goto',
  'inline',
  'noinline',
  'volatile',
  'public',
  'static',
  'extern',
  'external',
  'interface',
  'long',
  'short',
  'double',
  'half',
  'fixed',
  'unsigned',
  'superp',
  'input',
  'output',
  'hvec2',
  'hvec3',
  'hvec4',
  'dvec2',
  'dvec3',
  'dvec4',
  'fvec2',
  'fvec3',
  'fvec4',
  'sampler3DRect',
  'filter',
  'image1D',
  'image2D',
  'image3D',
  'imageCube',
  'iimage1D',
  'iimage2D',
  'iimage3D',
  'iimageCube',
  'uimage1D',
  'uimage2D',
  'uimage3D',
  'uimageCube',
  'image1DArray',
  'image2DArray',
  'iimage1DArray',
  'iimage2DArray',
  'uimage1DArray',
  'uimage2DArray',
  'imageBuffer',
  'iimageBuffer',
  'uimageBuffer',
  'sampler1D',
  'sampler1DShadow',
  'sampler1DArray',
  'sampler1DArrayShadow',
  'isampler1D',
  'isampler1DArray',
  'usampler1D',
  'usampler1DArray',
  'sampler2DRect',
  'sampler2DRectShadow',
  'isampler2DRect',
  'usampler2DRect',
  'samplerBuffer',
  'isamplerBuffer',
  'usamplerBuffer',
  'sampler2DMS',
  'isampler2DMS',
  'usampler2DMS',
  'sampler2DMSArray',
  'isampler2DMSArray',
  'usampler2DMSArray',
  'sizeof',
  'cast',
  'namespace',
  'using',

  // 3.5 Preprocessor
  '#define',
  '#undef',
  '#if',
  '#ifdef',
  '#ifndef',
  '#else',
  '#elif',
  '#endif',
  '#error',
  '#pragma',
  '#extension',
  '#version',
  '#line',

  // 3.5 Preprocessor - Operators
  'defined',

  // 3.5 Preprocessor - Macros
  '__LINE__',
  '__FILE__',
  '__VERSION__',
  'GL_ES',

  // 7.1 Vertex Shader Special Variables
  'gl_VertexID',
  'gl_InstanceID',
  'gl_Position',
  'gl_PointSize',

  // 7.2 Fragment Shader Special Variables
  'gl_FragCoord',
  'gl_FrontFacing',
  'gl_FragDepth',
  'gl_PointCoord',

  // 7.3 Built-in Constants
  'gl_MaxVertexAttribs',
  'gl_MaxVertexUniformVectors',
  'gl_MaxVertexOutputVectors',
  'gl_MaxFragmentInputVectors',
  'gl_MaxVertexTextureImageUnits',
  'gl_MaxCombinedTextureImageUnits',
  'gl_MaxTextureImageUnits',
  'gl_MaxFragmentUniformVectors',
  'gl_MaxDrawBuffers',
  'gl_MinProgramTexelOffset',
  'gl_MaxProgramTexelOffset',

  // 7.4 Built-in Uniform State
  'gl_DepthRangeParameters',
  'gl_DepthRange',

  // 8.1 Angle and Trigonometry Functions
  'radians',
  'degrees',
  'sin',
  'cos',
  'tan',
  'asin',
  'acos',
  'atan',
  'sinh',
  'cosh',
  'tanh',
  'asinh',
  'acosh',
  'atanh',

  // 8.2 Exponential Functions
  'pow',
  'exp',
  'log',
  'exp2',
  'log2',
  'sqrt',
  'inversesqrt',

  // 8.3 Common Functions
  'abs',
  'sign',
  'floor',
  'trunc',
  'round',
  'roundEven',
  'ceil',
  'fract',
  'mod',
  'modf',
  'min',
  'max',
  'clamp',
  'mix',
  'step',
  'smoothstep',
  'isnan',
  'isinf',
  'floatBitsToInt',
  'floatBitsToUint',
  'intBitsToFloat',
  'uintBitsToFloat',

  // 8.4 Floating-Point Pack and Unpack Functions
  'packSnorm2x16',
  'unpackSnorm2x16',
  'packUnorm2x16',
  'unpackUnorm2x16',
  'packHalf2x16',
  'unpackHalf2x16',

  // 8.5 Geometric Functions
  'length',
  'distance',
  'dot',
  'cross',
  'normalize',
  'faceforward',
  'reflect',
  'refract',
  'matrixCompMult',
  'outerProduct',
  'transpose',
  'determinant',
  'inverse',

  // 8.7 Vector Relational Functions
  'lessThan',
  'lessThanEqual',
  'greaterThan',
  'greaterThanEqual',
  'equal',
  'notEqual',
  'any',
  'all',
  'not',

  // 8.8 Texture Lookup Functions
  'textureSize',
  'texture',
  'textureProj',
  'textureLod',
  'textureOffset',
  'texelFetch',
  'texelFetchOffset',
  'textureProjOffset',
  'textureLodOffset',
  'textureProjLod',
  'textureProjLodOffset',
  'textureGrad',
  'textureGradOffset',
  'textureProjGrad',
  'textureProjGradOffset',

  // 8.9 Fragment Processing Functions
  'dFdx',
  'dFdy',
  'fwidth',

  // Additional directives
  '#include',

  // WEBGL_multi_draw https://registry.khronos.org/webgl/extensions/WEBGL_multi_draw
  'gl_DrawID', // vertex only

  // OVR_multiview2 https://registry.khronos.org/webgl/extensions/OVR_multiview2
  // OCULUS_multiview https://github.com/KhronosGroup/WebGL/issues/2912
  'gl_ViewID_OVR',
  'GL_OVR_multiview2',
]

const SYMBOLS = [
  // Comments
  '//',
  '/*',
  '*/',

  // Punctuation
  ':',
  ',',
  '.',
  '{',
  '[',
  '(',
  '?',
  '}',
  ']',
  ')',
  ';',

  // Unary
  '~',
  '--',
  '++',
  '!',

  // Binary
  '&',
  '|',
  '^',
  '/',
  '==',
  '>',
  '>=',
  '<',
  '<=',
  '&&',
  '||',
  '^^',
  '-',
  '*',
  '!=',
  '+',
  '%',
  '<<',
  '>>',

  // Binary assignment
  '=',
  '+=',
  '&=',
  '|=',
  '^=',
  '/=',
  '*=',
  '%=',
  '<<=',
  '>>=',
  '-=',
]

// GLSL 3.2 Character Set, 5.1 Operators
export const GLSL_SYMBOLS = [
  ...SYMBOLS,
  // Preprocessor
  '#',
  // Line continuation
  '\\',
]

// WGSL 15.3 Syntactic Tokens
export const WGSL_SYMBOLS = [
  ...SYMBOLS,
  // Function arrow
  '->',
  // Attribute
  '@',
]
